@@ -1,6 +1,6 @@
 import warnings
 
-from typing import Optional, Tuple
+from typing import Any, Callable, List, Optional, Tuple, Union
 
 import numpy as np
 import scipy.stats
@@ -43,6 +43,7 @@ class _EntropyCoder:
             encoder = range_coder.RangeEncoder()
             decoder = range_coder.RangeDecoder()
 
+        self.name = method
         self._encoder = encoder
         self._decoder = decoder
 
@@ -59,12 +60,16 @@ def default_entropy_coder():
     return get_entropy_coder()
 
 
-def pmf_to_quantized_cdf(pmf, precision=16):
+def pmf_to_quantized_cdf(pmf: Tensor, precision: int = 16) -> Tensor:
     cdf = _pmf_to_quantized_cdf(pmf.tolist(), precision)
     cdf = torch.IntTensor(cdf)
     return cdf
 
 
+def _forward(self, *args: Any) -> Any:
+    raise NotImplementedError()
+
+
 class EntropyModel(nn.Module):
     r"""Entropy model base class.
 
@@ -76,7 +81,10 @@ class EntropyModel(nn.Module):
     """
 
     def __init__(
-        self, likelihood_bound=1e-9, entropy_coder=None, entropy_coder_precision=16
+        self,
+        likelihood_bound: float = 1e-9,
+        entropy_coder: Optional[str] = None,
+        entropy_coder_precision: int = 16,
     ):
         super().__init__()
 
@@ -94,6 +102,15 @@ class EntropyModel(nn.Module):
         self.register_buffer("_quantized_cdf", torch.IntTensor())
         self.register_buffer("_cdf_length", torch.IntTensor())
 
+    def __getstate__(self):
+        attributes = self.__dict__.copy()
+        attributes["entropy_coder"] = self.entropy_coder.name
+        return attributes
+
+    def __setstate__(self, state):
+        self.__dict__ = state
+        self.entropy_coder = _EntropyCoder(self.__dict__.pop("entropy_coder"))
+
     @property
     def offset(self):
         return self._offset
@@ -106,18 +123,8 @@ class EntropyModel(nn.Module):
     def cdf_length(self):
         return self._cdf_length
 
-    def forward(self, *args):
-        raise NotImplementedError()
-
-    @torch.jit.unused
-    def _get_noise_cached(self, x):
-        # use simple caching method to avoid creating a new tensor every call
-        half = float(0.5)
-        if not hasattr(self, "_noise"):
-            setattr(self, "_noise", x.new(x.size()))
-        self._noise.resize_(x.size())
-        self._noise.uniform_(-half, half)
-        return self._noise
+    # See: https://github.com/python/mypy/issues/8795
+    forward: Callable[..., Any] = _forward
 
     def quantize(
         self, inputs: Tensor, mode: str, means: Optional[Tensor] = None
@@ -126,11 +133,8 @@ class EntropyModel(nn.Module):
             raise ValueError(f'Invalid quantization mode: "{mode}"')
 
         if mode == "noise":
-            if torch.jit.is_scripting():
-                half = float(0.5)
-                noise = torch.empty_like(inputs).uniform_(-half, half)
-            else:
-                noise = self._get_noise_cached(inputs)
+            half = float(0.5)
+            noise = torch.empty_like(inputs).uniform_(-half, half)
             inputs = inputs + noise
             return inputs
 
@@ -149,12 +153,14 @@ class EntropyModel(nn.Module):
         outputs = outputs.int()
         return outputs
 
-    def _quantize(self, inputs, mode, means=None):
+    def _quantize(
+        self, inputs: Tensor, mode: str, means: Optional[Tensor] = None
+    ) -> Tensor:
         warnings.warn("_quantize is deprecated. Use quantize instead.")
         return self.quantize(inputs, mode, means)
 
     @staticmethod
-    def dequantize(inputs, means=None):
+    def dequantize(inputs: Tensor, means: Optional[Tensor] = None) -> Tensor:
         if means is not None:
             outputs = inputs.type_as(means)
             outputs += means
@@ -163,7 +169,7 @@ class EntropyModel(nn.Module):
         return outputs
 
     @classmethod
-    def _dequantize(cls, inputs, means=None):
+    def _dequantize(cls, inputs: Tensor, means: Optional[Tensor] = None) -> Tensor:
         warnings.warn("_dequantize. Use dequantize instead.")
         return cls.dequantize(inputs, means)
 
@@ -209,8 +215,10 @@ class EntropyModel(nn.Module):
         """
         symbols = self.quantize(inputs, "symbols", means)
 
-        if len(inputs.size()) != 4:
-            raise ValueError("Invalid `inputs` size. Expected a 4-D tensor.")
+        if len(inputs.size()) < 2:
+            raise ValueError(
+                "Invalid `inputs` size. Expected a tensor with at least 2 dimensions."
+            )
 
         if inputs.size() != indexes.size():
             raise ValueError("`inputs` and `indexes` should have the same size.")
@@ -247,20 +255,22 @@ class EntropyModel(nn.Module):
         if not len(strings) == indexes.size(0):
             raise ValueError("Invalid strings or indexes parameters")
 
-        if len(indexes.size()) != 4:
-            raise ValueError("Invalid `indexes` size. Expected a 4-D tensor.")
+        if len(indexes.size()) < 2:
+            raise ValueError(
+                "Invalid `indexes` size. Expected a tensor with at least 2 dimensions."
+            )
 
         self._check_cdf_size()
         self._check_cdf_length()
         self._check_offsets_size()
 
         if means is not None:
-            if means.size()[:-2] != indexes.size()[:-2]:
+            if means.size()[:2] != indexes.size()[:2]:
                 raise ValueError("Invalid means or indexes parameters")
-            if means.size() != indexes.size() and (
-                means.size(2) != 1 or means.size(3) != 1
-            ):
-                raise ValueError("Invalid means parameters")
+            if means.size() != indexes.size():
+                for i in range(2, len(indexes.size())):
+                    if means.size(i) != 1:
+                        raise ValueError("Invalid means parameters")
 
         cdf = self._quantized_cdf
         outputs = cdf.new(indexes.size())
@@ -290,14 +300,16 @@ class EntropyBottleneck(EntropyModel):
     for an introduction.
     """
 
+    _offset: Tensor
+
     def __init__(
         self,
-        channels,
-        *args,
-        tail_mass=1e-9,
-        init_scale=10,
-        filters=(3, 3, 3, 3),
-        **kwargs,
+        channels: int,
+        *args: Any,
+        tail_mass: float = 1e-9,
+        init_scale: float = 10,
+        filters: Tuple[int, ...] = (3, 3, 3, 3),
+        **kwargs: Any,
     ):
         super().__init__(*args, **kwargs)
 
@@ -333,11 +345,11 @@ class EntropyBottleneck(EntropyModel):
         target = np.log(2 / self.tail_mass - 1)
         self.register_buffer("target", torch.Tensor([-target, 0, target]))
 
-    def _get_medians(self):
+    def _get_medians(self) -> Tensor:
         medians = self.quantiles[:, :, 1:2]
         return medians
 
-    def update(self, force=False):
+    def update(self, force: bool = False) -> bool:
         # Check if we need to update the bottleneck parameters, the offsets are
         # only computed and stored when the conditonal model is update()'d.
         if self._offset.numel() > 0 and not force:  # pylint: disable=E0203
@@ -358,7 +370,7 @@ class EntropyBottleneck(EntropyModel):
         pmf_start = medians - minima
         pmf_length = maxima + minima + 1
 
-        max_length = pmf_length.max()
+        max_length = pmf_length.max().item()
         device = pmf_start.device
         samples = torch.arange(max_length, device=device)
 
@@ -379,12 +391,12 @@ class EntropyBottleneck(EntropyModel):
         self._cdf_length = pmf_length + 2
         return True
 
-    def loss(self):
+    def loss(self) -> Tensor:
         logits = self._logits_cumulative(self.quantiles, stop_gradient=True)
         loss = torch.abs(logits - self.target).sum()
         return loss
 
-    def _logits_cumulative(self, inputs, stop_gradient):
+    def _logits_cumulative(self, inputs: Tensor, stop_gradient: bool) -> Tensor:
         # TorchScript not yet working (nn.Mmodule indexing not supported)
         logits = inputs
         for i in range(len(self.filters) + 1):
@@ -406,7 +418,7 @@ class EntropyBottleneck(EntropyModel):
         return logits
 
     @torch.jit.unused
-    def _likelihood(self, inputs):
+    def _likelihood(self, inputs: Tensor) -> Tensor:
         half = float(0.5)
         v0 = inputs - half
         v1 = inputs + half
@@ -419,11 +431,25 @@ class EntropyBottleneck(EntropyModel):
         )
         return likelihood
 
-    def forward(self, x: Tensor, training: Optional[bool] = None):
+    def forward(
+        self, x: Tensor, training: Optional[bool] = None
+    ) -> Tuple[Tensor, Tensor]:
         if training is None:
             training = self.training
-        # Convert to (channels, ... , batch) format
-        x = x.permute(1, 2, 3, 0).contiguous()
+
+        if not torch.jit.is_scripting():
+            # x from B x C x ... to C x B x ...
+            perm = np.arange(len(x.shape))
+            perm[0], perm[1] = perm[1], perm[0]
+            # Compute inverse permutation
+            inv_perm = np.arange(len(x.shape))[np.argsort(perm)]
+        else:
+            # TorchScript in 2D for static inference
+            # Convert to (channels, ... , batch) format
+            perm = (1, 2, 3, 0)
+            inv_perm = (3, 0, 1, 2)
+
+        x = x.permute(*perm).contiguous()
         shape = x.size()
         values = x.reshape(x.size(0), 1, -1)
 
@@ -443,29 +469,43 @@ class EntropyBottleneck(EntropyModel):
 
         # Convert back to input tensor shape
         outputs = outputs.reshape(shape)
-        outputs = outputs.permute(3, 0, 1, 2).contiguous()
+        outputs = outputs.permute(*inv_perm).contiguous()
 
         likelihood = likelihood.reshape(shape)
-        likelihood = likelihood.permute(3, 0, 1, 2).contiguous()
+        likelihood = likelihood.permute(*inv_perm).contiguous()
 
         return outputs, likelihood
 
     @staticmethod
     def _build_indexes(size):
-        N, C, H, W = size
-        indexes = torch.arange(C).view(1, -1, 1, 1)
+        dims = len(size)
+        N = size[0]
+        C = size[1]
+
+        view_dims = np.ones((dims,), dtype=np.int64)
+        view_dims[1] = -1
+        indexes = torch.arange(C).view(*view_dims)
         indexes = indexes.int()
-        return indexes.repeat(N, 1, H, W)
+
+        return indexes.repeat(N, 1, *size[2:])
+
+    @staticmethod
+    def _extend_ndims(tensor, n):
+        return tensor.reshape(-1, *([1] * n)) if n > 0 else tensor.reshape(-1)
 
     def compress(self, x):
         indexes = self._build_indexes(x.size())
-        medians = self._get_medians().detach().expand(x.size(0), -1, 1, 1)
+        medians = self._get_medians().detach()
+        spatial_dims = len(x.size()) - 2
+        medians = self._extend_ndims(medians, spatial_dims)
+        medians = medians.expand(x.size(0), *([-1] * (spatial_dims + 1)))
         return super().compress(x, indexes, medians)
 
     def decompress(self, strings, size):
-        output_size = (len(strings), self._quantized_cdf.size(0), size[0], size[1])
+        output_size = (len(strings), self._quantized_cdf.size(0), *size)
         indexes = self._build_indexes(output_size).to(self._quantized_cdf.device)
-        medians = self._get_medians().detach().expand(len(strings), -1, 1, 1)
+        medians = self._extend_ndims(self._get_medians().detach(), len(size))
+        medians = medians.expand(len(strings), *([-1] * (len(size) + 1)))
         return super().decompress(strings, indexes, medians)
 
 
@@ -480,7 +520,14 @@ class GaussianConditional(EntropyModel):
     for more information.
     """
 
-    def __init__(self, scale_table, *args, scale_bound=0.11, tail_mass=1e-9, **kwargs):
+    def __init__(
+        self,
+        scale_table: Optional[Union[List, Tuple]],
+        *args: Any,
+        scale_bound: float = 0.11,
+        tail_mass: float = 1e-9,
+        **kwargs: Any,
+    ):
         super().__init__(*args, **kwargs)
 
         if not isinstance(scale_table, (type(None), list, tuple)):
@@ -496,11 +543,10 @@ class GaussianConditional(EntropyModel):
 
         self.tail_mass = float(tail_mass)
         if scale_bound is None and scale_table:
-            self.lower_bound_scale = LowerBound(self.scale_table[0])
-        elif scale_bound > 0:
-            self.lower_bound_scale = LowerBound(scale_bound)
-        else:
+            scale_bound = self.scale_table[0]
+        if scale_bound <= 0:
             raise ValueError("Invalid parameters")
+        self.lower_bound_scale = LowerBound(scale_bound)
 
         self.register_buffer(
             "scale_table",
@@ -516,8 +562,7 @@ class GaussianConditional(EntropyModel):
     def _prepare_scale_table(scale_table):
         return torch.Tensor(tuple(float(s) for s in scale_table))
 
-    def _standardized_cumulative(self, inputs):
-        # type: (Tensor) -> Tensor
+    def _standardized_cumulative(self, inputs: Tensor) -> Tensor:
         half = float(0.5)
         const = float(-(2 ** -0.5))
         # Using the complementary error function maximizes numerical precision.
@@ -563,8 +608,9 @@ class GaussianConditional(EntropyModel):
         self._offset = -pmf_center
         self._cdf_length = pmf_length + 2
 
-    def _likelihood(self, inputs, scales, means=None):
-        # type: (Tensor, Tensor, Optional[Tensor]) -> Tensor
+    def _likelihood(
+        self, inputs: Tensor, scales: Tensor, means: Optional[Tensor] = None
+    ) -> Tensor:
         half = float(0.5)
 
         if means is not None:
@@ -596,7 +642,7 @@ class GaussianConditional(EntropyModel):
             likelihood = self.likelihood_lower_bound(likelihood)
         return outputs, likelihood
 
-    def build_indexes(self, scales):
+    def build_indexes(self, scales: Tensor) -> Tensor:
         scales = self.lower_bound_scale(scales)
         indexes = scales.new_full(scales.size(), len(self.scale_table) - 1).int()
         for s in self.scale_table[:-1]:
